/*------------------------------------------------------------------------------
* sbspos.c : single point positioning with sbas dgps correction
*
*          Copyright (C) 2007-2009 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1.1 $ $Date: 2008/07/17 22:04:02 $
* history : 2007/10/22 1.0 new
*           2008/01/24 1.1 support rtklib 2.2.0 api change
*           2009/11/12 1.2 support rtklib 2.3.0 api change
*           2010/01/22 1.3 support ems for sbas correction file
*-----------------------------------------------------------------------------*/
#include "rtklib.h"

static const char rcsid[]="$Id: sbspos.c,v 1.1 2008/07/17 22:04:02 ttaka Exp $";

#define PROGNAME    "sbspos ver.2.1"
#define SIGPR       0.3         /* standard dev of pseudorange obs (m) */
#define SIGDP       0.01        /* standard dev of doppler obs (m/s) */
#define SIGACC      0.1         /* standard dev of accelation (m/s^2) */
#define TIMEOUTF    30.0        /* smoothing filter timeout (s) */

/* help text -----------------------------------------------------------------*/
static const char *help[]={
"",
" usage: sbspos [option ...] file file [...]",
"",
" Single point positioning with SBAS DGPS corrections. Files shall include",
" receiver RINEX OBS file, NAV file and SBAS message log file (.sbs or .ems).",
" Command options are as follows. ([]:default)",
"",
" -o output output file [stdout]",
" -b        SBAS satellite prn number [129]",
" -m mask   elevation mask angle (deg) [15]",
" -c mask   snr mask (dbHz) [0]",
" -p        single point positioing without SBAS DGPS corrections [off]",
" -l        apply SBAS long term corrections [all]",
" -i        apply SBAS ionospheric corrections [all]",
" -f        apply SBAS fast corrections [all]",
" -r        apply SBAS ranging [all]",
" -s        apply doppler smoothing [off]",
};
/* print help ----------------------------------------------------------------*/
static void printhelp(void)
{
    int i;
    for (i=0;i<sizeof(help)/sizeof(*help);i++) fprintf(stderr,"%s\n",help[i]);
    exit(0);
}
/* doppler smoothing ---------------------------------------------------------*/
static void dsmooth(const obsd_t *obs, int n, obsd_t *sobs)
{
    const double lam=CLIGHT/FREQ1;
    static obsd_t obs_[MAXOBS];
    static double x[MAXOBS],v[MAXOBS];
    static int np=0;
    int i,j;
    double t,dp,dd,vd,k;
    for (i=0;i<n;i++) {
        sobs[i]=obs[i];
        for (j=0;j<np;j++) if (obs_[j].sat==obs[i].sat) break;
        
        if (j>=np) {
            x[j]=obs[i].P[0];
            v[j]=SIGPR*SIGPR;
            if (j>=np) np=j+1;
        }
        else {
            t=timediff(obs[i].time,obs_[j].time);
            dp=obs[i].P[0]-obs_[j].P[0];
            dd=-lam*(obs[i].D[0]+obs_[j].D[0])/2.0*t;
            vd=SIGDP*SIGDP*t+SIGACC*SIGACC*t*t/2.0;
            
            if (fabs(t)<=TIMEOUTF&&(dp-dd)*(dp-dd)<=9.0*(v[j]+vd)) {
                /* temporal update */
                x[j]+=dd; v[j]+=vd;
                
                /* measurement update */
                k=v[j]/(v[j]+SIGPR*SIGPR);
                x[j]+=k*(obs[i].P[0]-x[j]);
                v[j]-=k*v[j];
                sobs[i].P[0]=x[j];
            }
            else {
                x[j]=obs[i].P[0];
                v[j]=SIGPR*SIGPR;
            }
        }
        obs_[j]=obs[i];
    }
}
/* sbas positioning ----------------------------------------------------------*/
static void sbaspos(FILE *fp, const obs_t *obs, nav_t *nav,
                    const sbsmsg_t *msg, int nmsg, double elmin, double snrmin,
                    int opt, int smooth)
{
    int i,j,k,m,ns,q,vsat[MAXOBS],corr;
    char s[64];
    double rr[3]={0},rb[3]={0},Q[9],dtr,azel[MAXOBS*2],res[MAXOBS];
    gtime_t time;
    sbsstat_t stat={{0}};
    obsd_t sobs[MAXOBS];
    
    for (i=k=0;i<obs->n;i=j) {
        for (j=i;j<obs->n;j++) {
            if (timediff(obs->data[j].time,obs->data[i].time)>0.0) break;
        }
        time=obs->data[i].time;
        for (;k<nmsg;k++) {
            if (timediff(gpst2time(msg[k].week,msg[k].tow),time)>DTTOL) break;
            sbsupdatestat(msg+k,&stat,nav);
        }
        /* doppler smoothing */
        if (smooth) dsmooth(obs->data+i,j-i,sobs);
        else for (m=i;m<j;m++) sobs[m-i]=obs->data[m];
        
        /* sbas positioning */
        if (opt<16&&(ns=sbspntpos(sobs,j-i,nav,&stat,elmin,snrmin,opt,rr,Q,
                     &dtr,azel,vsat,&corr))>0) {
            q=corr?4:5;
        }
        /* point positioning */
        else if ((ns=pntpos(sobs,j-i,nav,elmin,snrmin,rr,Q,&dtr,azel,vsat,res))>0) {
            q=5;
        }
        else continue;
        outsol(fp,timeadd(time,-dtr),rr,Q,rb,q,ns,0.0,0.0);
        
        time2str(time,s,0);
        fprintf(stderr," %s : NS=%2d NV=%2d Q=%d\r",s,j-i,ns,q);
    }
}
/* main ----------------------------------------------------------------------*/
int main(int argc, char **argv)
{
    int i,n=0,nmsg=0,opt=0,smooth=0,sat=129;
    char *files[32]={0},*sfile="",*ofile="",*p;
    double elmin=15.0,snrmin=0.0;
    FILE *fp=stdout;
    nav_t nav={0};
    obs_t obs={0};
    sbsmsg_t *msg=NULL;
    seph_t seph[MAXPRNSBS-MINPRNSBS+1]={{0}};
    
    nav.seph=seph;
    nav.ns=nav.nsmax=MAXPRNSBS-MINPRNSBS+1;
    
    for (i=1;i<argc;i++) {
        if      (!strcmp(argv[i],"-p")) opt=16;
        else if (!strcmp(argv[i],"-l")) opt|=SBSOPT_LCORR;
        else if (!strcmp(argv[i],"-i")) opt|=SBSOPT_ICORR;
        else if (!strcmp(argv[i],"-f")) opt|=SBSOPT_FCORR;
        else if (!strcmp(argv[i],"-r")) opt|=SBSOPT_RANGE;
        else if (!strcmp(argv[i],"-s")) smooth=1;
        else if (!strcmp(argv[i],"-b")&&i+1<argc) sat=atoi(argv[++i]);
        else if (!strcmp(argv[i],"-m")&&i+1<argc) elmin=atof(argv[++i]);
        else if (!strcmp(argv[i],"-c")&&i+1<argc) snrmin=atof(argv[++i]);
        else if (!strcmp(argv[i],"-o")&&i+1<argc) ofile=argv[++i];
        else if (!strncmp(argv[i],"-",1)) printhelp();
        else if ((p=strrchr(argv[i],'.'))&&!strcmp(p,".sbs")) sfile=argv[i];
        else if ((p=strrchr(argv[i],'.'))&&!strcmp(p,".ems")) sfile=argv[i];
        else files[n++]=argv[i];
    }
    if (opt!=16&&(nmsg=sbsreadmsg(sfile,sat,&msg))<=0) {
        fprintf(stderr,"no sbas messages : %s\n",sfile);
        return -1;
    }
    if (readrnx(files,n,&obs,&nav)<=0) {
        fprintf(stderr,"no observation data\n");
        return -1;
    }
    if (nav.n<=0) {
        fprintf(stderr,"no navigation message\n");
        return -1;
    }
    if (ofile[0]!='\0'&&!(fp=fopen(ofile,"w"))) {
        fprintf(stderr,"output file open error : %s\n",ofile);
        return -1;
    }
    fprintf(fp,"%% program   : %s\n",PROGNAME);
    for (i=0;i<n;i++) {
        fprintf(fp,"%% inp file  : %s",files[i]);
    }
    fprintf(fp,"%% sbas file : %s\n",sfile);
    fprintf(fp,"%% sbas sat  : PRN%d\n",sat);
    fprintf(fp,"%% elev mask : %.1f deg\n",elmin);
    outsolhead(fp);
    sbaspos(fp,&obs,&nav,msg,nmsg,elmin*D2R,snrmin,opt==0?15:opt,smooth);
    fclose(fp);
    return 0;
}
